/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "fonl_driver.hh"
#include "loc_list.hh"
#include "src_update.hh"
#include "update_list.hh"
#include "helpers.hh"
#include <stdexcept>

#include "ldas/ldasconfig.hh"
#ifndef LDAS_VERSION_NUMBER
#define LDAS_VERSION_NUMBER 109000
#endif

#if LDAS_VERSION_NUMBER < 109360
#error "backward compatibility broken before framecpp-1.9.360"
#endif

#if LDAS_VERSION_NUMBER > 119010
#define FCPP_SHARED_PTRS 1
#if LDAS_VERSION_NUMBER >=200000
#include "ldastoolsal/SharedPtr.hh"
using LDASTools::AL::SharedPtr;
#else
#include "general/SharedPtr.hh"
using General::SharedPtr;
#endif
#else
#error "A release of FrameCPP with LDAS_VERSION_NUMBER > 119010 is required"
#endif

#include "framecpp/IFrameStream.hh"

#include <fstream>
#include <iostream>
#include <unistd.h>

using namespace std;
using namespace sends;

//======================================  Constructor
fonl_driver::fonl_driver(const std::string& dname) 
    : mUpdate(NULL)
{
    set_name(dname);
}

//======================================  Destructor
fonl_driver::~fonl_driver(void) {
}

//======================================  Get a pointer to the updater
bool
fonl_driver::find_update(void) {
    if (mUpdate == NULL && !mUpdName.empty()) {
	mUpdate = UpdateList.find(mUpdName);
    }
    return mUpdate != NULL;
}

//======================================  Get the current online gps time
gps_type
fonl_driver::online_gps(void) {
    gps_type gps = 0;

    //----------------------------------  Get current time if updater available
    if (find_update()) {
	gps = mUpdate->current_time();
	while (!gps) {
	    mUpdate->wait();
	    gps = mUpdate->current_time();
	}
    }

    //----------------------------------  Otherwise return gps = 0
    else {
	cerr << name() << ": Updater: " << mUpdName << " not available." 
	     << endl;
    }

    //----------------------------------  Return gps time
    return gps;
}

//======================================  Read data into segments
int
fonl_driver::read_data(const loc_list& ll, gps_type start, 
		       gps_type stop, const channel_vect& v) {

    //----------------------------------  Define a segment number list.
    segment_vect segv;
    if (!find_update() ) {
	throw runtime_error("Can't read online frames with undefined updater");
    }

    //----------------------------------  Loop over frames
    int rc = 0;
    while (start < stop) {
	//------------------------------  Make up a maximum wait time
	frame_loc fl;
	ll.find_back(fl);
	double max_wait = 120.0;
	double dt = fl.duration();
	if (dt > 60.0) max_wait = dt + 60.0;

	//------------------------------  Wait for the data to appear.
	if (!mUpdate->wait(start, max_wait)) {
	    rc = -1;
	    break;
	}

	//------------------------------  Get the frame file.
	ll.find_back(fl);
	string file = fl.file_path(start);
	if (file.empty()) {
	    cerr << "fonl_driver: Can't find frame for gps= " << start 
		 << " in " << loc_list::arch_name(ll.archive()) << ":" 
		 << fl.prefix() << endl;
	    rc = -1;
	    break;
	} else if (access(file.c_str(), R_OK) != 0) {
	    cerr << "fonl_driver: Can't find frame: " << file << endl;
	    rc = -1;
	    break;
	}
	gps_type end_frame   = fl.frame_end(start);
	gps_type start_frame = fl.frame_start(start);

	if (!get_segments(start_frame, end_frame, v, segv)) {
	    start = end_frame;
	    continue;
	}

	try {
	    //--------------------------  Open the frame
	    if (debug() >= 2) cerr << "Read data from frame: " << file << endl;
	    SharedPtr< FrameCPP::Common::FrameBuffer<filebuf> > 
	    frbuf( new FrameCPP::Common::FrameBuffer<filebuf>(ios::in) );
	    if (! frbuf.get()->open(file.c_str(), ios::in) ) 
		throw runtime_error( string("Unable to open file: ") + file);
	    
		// ensure that we always call the close method
		auto_closer< FrameCPP::Common::FrameBuffer<filebuf> > _closer(*frbuf);
	    
	    FrameCPP::IFrameStream frdr(false, frbuf.get());

	    //--------------------------  Loop over frames, read data.
	    int nFrames = frdr.GetTOC()->nFrame();
	    for (int iFrame=0; iFrame < nFrames; ++iFrame) {
		if (read_frame(frdr, segv, iFrame)) rc = -1;
	    }

	    //--------------------------  Finish processing
	    start = end_frame;

	//------------------------------  On error, release segments and 
	//                                rethrow exception
	} catch (...) {
	    release_segments(segv);
	    throw;
	}
	release_segments(segv);
    }
    return rc;
}

//======================================  Set the driver name
void
fonl_driver::set_name(const std::string& nm) {
    mName  = "fonl:";
    mName += nm;
}

//======================================  Set the updater
void
fonl_driver::set_update(const std::string& upd) {
    mUpdName = upd;
    find_update();
}
